import { z, ZodArray, ZodDiscriminatedUnion, ZodUndefined } from 'zod';
import { isCustomScalarSchema } from './custom-scalar.js';
import { isFragmentsSchema, isObjectOrListSchema, isStrictObjectSchema, isUnionOrListSchema, unwrapFieldSchema, unwrapFieldSchemaChain } from './query-schema.js';
import { normalizeParameterList } from './values/parameter-list.js';
import { ensureValidVariableCorrelations } from './variable-correlations.js';
import { serializeVariableDefinitions } from './variable-definition.js';
function withTrailingSpace(value) {
    if (value.length === 0) {
        return value;
    }
    return `${value} `;
}
function unwrapFromArraySchema(predicate, schema) {
    const elementSchema = unwrapFieldSchema(schema.element);
    if (predicate(elementSchema)) {
        return elementSchema;
    }
    return null;
}
function unwrapFromTupleSchema(predicate, schema) {
    const [firstElementSchema] = schema.items;
    const unwrappedElementSchema = unwrapFieldSchema(firstElementSchema);
    if (predicate(unwrappedElementSchema)) {
        return unwrappedElementSchema;
    }
    return null;
}
function getObjectSchema(schema) {
    if (isStrictObjectSchema(schema)) {
        return schema;
    }
    if (schema instanceof ZodArray) {
        return unwrapFromArraySchema(isStrictObjectSchema, schema);
    }
    return unwrapFromTupleSchema(isStrictObjectSchema, schema);
}
function getUnionSchema(schema) {
    if (schema instanceof ZodDiscriminatedUnion) {
        return schema;
    }
    if (schema instanceof ZodArray) {
        return unwrapFromArraySchema(isFragmentsSchema, schema);
    }
    return unwrapFromTupleSchema(isFragmentsSchema, schema);
}
export function createQueryBuilder() {
    const fieldOptionsRegistry = new WeakMap();
    function getFieldOptionsForSchema(schema) {
        const unwrappingResult = unwrapFieldSchemaChain(schema);
        const queue = [...unwrappingResult.wrapperElements, unwrappingResult.unwrappedSchema];
        for (const currentSchema of queue) {
            const fieldOptions = fieldOptionsRegistry.get(currentSchema);
            if (fieldOptions !== undefined) {
                return fieldOptions;
            }
        }
        return {};
    }
    function serializedFieldSelector(fieldName, fieldSchema) {
        const fieldOptions = getFieldOptionsForSchema(fieldSchema);
        const { aliasFor, parameters } = fieldOptions;
        const selectorName = aliasFor === undefined ? fieldName : `${fieldName}: ${aliasFor}`;
        if (parameters !== undefined) {
            const normalizedParameterList = normalizeParameterList(parameters);
            return {
                serializedValue: `${selectorName}${normalizedParameterList.serializedValue}`,
                referencedVariables: normalizedParameterList.referencedVariables
            };
        }
        return {
            serializedValue: selectorName,
            referencedVariables: new Set()
        };
    }
    function serializeObjectSchema(objectSchema) {
        const entries = Object.entries(objectSchema.shape);
        let referencedVariables = new Set();
        const serializedEntries = [];
        for (const [fieldName, fieldSchema] of entries) {
            // eslint-disable-next-line @typescript-eslint/no-use-before-define -- recursion
            const serializedField = serializeFieldSchema(fieldName, fieldSchema);
            referencedVariables = new Set([...referencedVariables, ...serializedField.referencedVariables]);
            serializedEntries.push(serializedField.serializedValue);
        }
        return {
            serializedValue: ` { ${serializedEntries.join(', ')} }`,
            referencedVariables
        };
    }
    function serializeFragments(unionOptions) {
        let referencedVariables = new Set();
        const serializedFragments = [];
        for (const [fragmentName, fragmentSchema] of unionOptions.entries()) {
            const serializedFragment = serializeObjectSchema(fragmentSchema);
            referencedVariables = new Set([...referencedVariables, ...serializedFragment.referencedVariables]);
            serializedFragments.push(`... on ${fragmentName}${serializedFragment.serializedValue}`);
        }
        return {
            serializedValue: ` { ${serializedFragments.join(', ')} }`,
            referencedVariables
        };
    }
    function combineFieldSelectorAndFieldBody(fieldSelector, fieldBody) {
        return {
            serializedValue: `${fieldSelector.serializedValue}${fieldBody.serializedValue}`,
            referencedVariables: new Set([
                ...fieldSelector.referencedVariables,
                ...fieldBody.referencedVariables
            ])
        };
    }
    // eslint-disable-next-line max-statements, complexity -- no good idea how to refactor at the moment
    function serializeFieldSchema(fieldName, fieldSchema) {
        const fieldSelector = serializedFieldSelector(fieldName, fieldSchema);
        if (isCustomScalarSchema(fieldSchema)) {
            return fieldSelector;
        }
        const unwrappedSchema = unwrapFieldSchema(fieldSchema);
        if (unwrappedSchema instanceof ZodUndefined) {
            return { serializedValue: '', referencedVariables: new Set() };
        }
        if (isObjectOrListSchema(unwrappedSchema)) {
            const objectSchema = getObjectSchema(unwrappedSchema);
            if (objectSchema !== null) {
                return combineFieldSelectorAndFieldBody(fieldSelector, serializeObjectSchema(objectSchema));
            }
        }
        if (isUnionOrListSchema(unwrappedSchema)) {
            const unionSchema = getUnionSchema(unwrappedSchema);
            if (unionSchema !== null) {
                return combineFieldSelectorAndFieldBody(fieldSelector, serializeFragments(unionSchema.optionsMap));
            }
        }
        return fieldSelector;
    }
    // eslint-disable-next-line max-statements -- no idea right now to make this smaller
    function buildDocument(documentType, schema, options) {
        let referencedVariables = new Set();
        const { variableDefinitions = {}, operationName = '' } = options;
        const bodyEntries = [];
        for (const [fieldName, fieldSchema] of Object.entries(schema.shape)) {
            const serializedField = serializeFieldSchema(fieldName, fieldSchema);
            if (serializedField.serializedValue.length > 0) {
                bodyEntries.push(serializedField.serializedValue);
            }
            referencedVariables = new Set([
                ...referencedVariables,
                ...serializedField.referencedVariables
            ]);
        }
        ensureValidVariableCorrelations(variableDefinitions, referencedVariables);
        const operationNameAndParams = withTrailingSpace(`${operationName}${serializeVariableDefinitions(variableDefinitions)}`);
        return `${documentType} ${operationNameAndParams}{ ${bodyEntries.join(', ')} }`;
    }
    return {
        registerFieldOptions(schema, options) {
            const newSchemaInstance = z.lazy(() => {
                return schema;
            });
            fieldOptionsRegistry.set(newSchemaInstance, options);
            return newSchemaInstance;
        },
        buildQuery(schema, options = {}) {
            return buildDocument('query', schema, options);
        },
        buildMutation(schema, options = {}) {
            return buildDocument('mutation', schema, options);
        }
    };
}
//# sourceMappingURL=builder.js.map